/*
 * Copyright (c) 2010, Noel Lemouel
 * All rights reserved.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>

#ifdef MACOSX
#include "SDL/include/SDL.h"
#else
#include "SDL.h"
#endif

#include "tga.h"

#define SCX 160
#define SCY 144

SDL_Surface *fb;

static int sc_open(FILE**, const char*, struct stat*);
static int so_create(FILE**, const char*);

float bg[SCX][SCY];

void fb_update(void)
{
	SDL_UpdateRect(fb, 0, 0, SCX, SCY);
}

void die(FILE* fd, unsigned char* mem, const char* str)
{
	printf("%s", str);

	free(mem);
	fclose(fd);

	exit(1);
}

int tile_extract(FILE* fd, int tx, int ty)
{
	unsigned int i, j;
	unsigned char x0;

	if (tx > 20 || ty > 18) return -1;

	fprintf(fd, "; tile x:%02d, y:%02d\n", tx, ty);

	tx=tx<<3;
	ty=ty<<3;

	for(j=0; j<8; j++)
	{
		fprintf(fd, ".db ");
		x0 = 0;
		for(i=0; i<8; i++)
		{
			x0 |= ((unsigned char)(bg[tx+i][ty+j]) & 0x01) << (7-i);
		}
		fprintf(fd, "$%02x,", ~x0 & 0xff);

		x0 = 0;
		for(i=0; i<8; i++)
		{
			x0 |= (((unsigned char)(bg[tx+i][ty+j]) >> 1) & 0x01) << (7-i);
		}
		fprintf(fd, "$%02x\n", ~x0 & 0xff);
	}
}

void bg_render(void)
{

	unsigned int pos_x;
	unsigned int pos_y;

	for (pos_y = 0; pos_y < SCY; pos_y++)
	{
		for (pos_x = 0; pos_x < SCX; pos_x++)
		{
			if(bg[pos_x][pos_y] == 0){
				((Uint32*)fb->pixels)[(pos_y*SCX) + pos_x] = 0xff4c3c1c;
			}else if (bg[pos_x][pos_y] == 1){
				((Uint32*)fb->pixels)[(pos_y*SCX) + pos_x] = 0xff987c3c;
			}else if (bg[pos_x][pos_y] == 2){
				((Uint32*)fb->pixels)[(pos_y*SCX) + pos_x] = 0xffdcb45c;
			}else if (bg[pos_x][pos_y] == 3){
				((Uint32*)fb->pixels)[(pos_y*SCX) + pos_x] = 0xfffce88c;
			}
		}
	}

}

void draw_box(unsigned int x, unsigned int y, unsigned int w, unsigned int h)
{
	unsigned int i;

	/* top */
	for(i=0; i<w; i++)
	{
		((Uint32*)fb->pixels)[(x+i)+(y*SCX)]= 0xffff0000;
	}

	/* bottom */
	for(i=0; i<w; i++)
	{
		((Uint32*)fb->pixels)[(x+i)+((h+y-1)*SCX)]= 0xffff0000;
	}

	/* left */
	for(i=0; i<h; i++)
	{
		((Uint32*)fb->pixels)[x+((y+i)*SCX)]= 0xffff0000;
	}

	/* right */
	for(i=0; i<h; i++)
	{
		((Uint32*)fb->pixels)[(x+w-1)+((y+i)*SCX)]= 0xffff0000;
	}

}

int main(int argc, char *argv[])
{

	tga_info_t tga_info_00;

	unsigned char* pxl_24;

	unsigned int pos_x;
	unsigned int pos_y;

	float r;
	float g;
	float b;
	float y;

	float pxl_00;
	float qe;

	FILE* fd0;
	FILE* fd1;
	struct stat info;

	unsigned char* fd0_data;

	unsigned int evnt_00;
	unsigned int evnt_01;

	/* open and load file */

	if(argc < 3){
		printf("expected 2 args [tga2bg input.tga output.inc]\n");
		return -1;
	}else{
		if( sc_open(&fd0, argv[1], &info) ) return -1;
	}

	fd0_data = (unsigned char*)malloc(info.st_size);
	if(!fd0_data){
		printf("error: [alloc. %d] failed\n", (int)info.st_size);
		return -1;
	}else{
		printf("[alloc. %d]\n", (int)info.st_size);
	}

	/* load fd0_data */

	fread(fd0_data, 1, info.st_size, fd0);

	/* ... */

	if (tga_info(fd0_data, &tga_info_00) != TGA_TYPE_TRUE_COLOR)
	{
		die(fd0, fd0_data, "invalid format: TGA_TYPE_TRUE_COLOR expected\n");
	}

	if ((tga_info_00.width != SCX) || (tga_info_00.height != SCY))
	{
		die(fd0, fd0_data, "invalid size: 160x144 expected\n");
	}

	/* ... */

	/* Initialize SDL's subsystems - in this case, only video. */
	if ( SDL_Init(SDL_INIT_VIDEO) < 0 )
	{
		fprintf(stderr, "Unable to init SDL: %s\n", SDL_GetError());
		exit(1);
	}

	/* Register SDL_Quit to be called at exit; makes sure things are
		cleaned up when we quit. */
	atexit(SDL_Quit);

	/* Attempt to create a SCXxSCY window with 32bit per pixels. */
	fb = SDL_SetVideoMode(SCX, SCY, 32, SDL_SWSURFACE);

	/* If we fail, return error. */
	if ( fb == NULL )
	{
		fprintf(stderr, "Unable to set %dx%d video: %s\n", SCX, SCY, SDL_GetError());
		exit(1);
	}

	SDL_WM_SetCaption ("tga2bg", NULL);

	pxl_24 = tga_info_00.base;

	for (pos_y = 0; pos_y < SCY; pos_y++)
	{
		for (pos_x = 0; pos_x < SCX; pos_x++)
		{
			/*
				((Uint32*)fb->pixels)[(pos_y*SCX) + pos_x] = (Uint32)(0x00) \
				| (pxl_24[2] << 16) \
				| (pxl_24[1] << 8) \
				| (pxl_24[0]);
			 */
			r = pxl_24[2] * 0.3;
			g = pxl_24[1] * 0.59;
			b = pxl_24[0] * 0.11;
			y = r + g + b;
			/* white threshold reduces noise introduced by the float conversion */
			if ((pxl_24[0]>0xfc) && (pxl_24[1]>0xfc) && (pxl_24[2]>0xfc)){
				bg[pos_x][pos_y] = (float)3.0;
			}else{
				bg[pos_x][pos_y] = (float)3/256*y;
			}
			pxl_24 += 3;
		}
	}

	for (pos_y = 0; pos_y < SCY; pos_y++)
	{
		for (pos_x = 0; pos_x < SCX; pos_x++)
		{

			pxl_00 = bg[pos_x][pos_y];
			if(pxl_00 < 0.5)
			{
				bg[pos_x][pos_y] = 0;
				qe = pxl_00 - bg[pos_x][pos_y];
			}else if((pxl_00 >= 0.5) && (pxl_00 < 1.5)){
				bg[pos_x][pos_y] = 1;
				qe = pxl_00 - bg[pos_x][pos_y];
			}else if((pxl_00 >= 1.5) && (pxl_00 < 2.5)){
				bg[pos_x][pos_y] = 2;
				qe = pxl_00 - bg[pos_x][pos_y];
			}else if(pxl_00 > 2.5){
				bg[pos_x][pos_y] = 3;
				qe = pxl_00 - bg[pos_x][pos_y];
			}

			if(pos_x < SCX-1)
				bg[pos_x+1][pos_y] += ((float)7/16) * qe;
			if((pos_x > 0) && (pos_y < SCY-1))
				bg[pos_x-1][pos_y+1] += ((float)3/16) * qe;
			if(pos_y < SCY-1)
				bg[pos_x][pos_y+1]  += ((float)5/16) * qe;
			if((pos_x < SCX-1) && (pos_y < SCY-1))
				bg[pos_x+1][pos_y+1] += ((float)1/16) * qe;
		}
	}

	bg_render();
	fb_update();

	evnt_00 = 0;
	evnt_01 = 0;

	while(1)
	{

		usleep(200000);

		/* Poll for events, and handle the ones we care about. */
		SDL_Event event;
		while (SDL_PollEvent(&event))
		{
			switch (event.type)
			{
				case SDL_KEYDOWN:
					break;
				case SDL_KEYUP:
					if (event.key.keysym.sym == SDLK_ESCAPE){
						free(fd0_data);
						fclose(fd0);
						return 0;
					}
					if (event.key.keysym.sym == ' '){
						if(!evnt_00){
							bg_render();
							draw_box(16, 8, 128, 128);
							fb_update();
							evnt_00 ++;
						}else{
							bg_render();
							fb_update();
							evnt_00 --;
						}
					}
					if (event.key.keysym.sym == 'w'){
						if(!evnt_01){
							if( so_create(&fd1, argv[2]) ) die(fd0, fd0_data, "invalid output filename\n");
								fprintf(fd1, "_bg_tile_data:\n");
								for(pos_y=1; pos_y<17; pos_y++){
									for(pos_x=2; pos_x<18; pos_x++)
										tile_extract(fd1, pos_x, pos_y);
								}
								fprintf(fd1, "\n");
							fclose(fd1);
							printf("tile data written to %s\n", argv[2]);
							evnt_01++;
						}
					}
					break;
				case SDL_QUIT:
					free(fd0_data);
					fclose(fd0);
					return(0);
			}
		}
	}
/*
	free(fd0_data);
	fclose(fd0);
	fclose(fd1);
*/
	return 0;	/* never reached */
}

static int sc_open(FILE** fd, const char* name, struct stat* info)
{
	if (stat(name, info) == -1)
	{
		if (errno == ENOENT)
		{
			(void) fprintf( stderr, "error: '%s' does not exist\r\n", name );
		}else{
			(void) fprintf( stderr, "error: stat() fails for '%s'\r\n", name );
		}
		return -1;
	}

	/* checking S_ISLNK(m) is not required since 'stat' follows sym. links */
	if (!S_ISREG(info->st_mode))
	{
		(void) fprintf( stderr, "error: '%s' is a not a regular file (or a sym. link)\r\n", name);
		return -1;
	}

	if ((*fd = fopen(name, "rb")) == NULL)
	{
		(void) fprintf( stderr, "error: '%s' open failed\r\n", name );
		return -1;
	}

	/* printf("size: %d\n", (int)info->st_size); */

	return 0;   /* no error */
}

static int so_create(FILE** fd, const char* name)
{
	struct stat so;
	unsigned char c;

    if (stat(name, &so) == -1)
    {
		if (errno == ENOENT)
        {
            /* does not exist => create it */
			if ((*fd = fopen(name, "wb")) == NULL)
      		{
         		(void) fprintf( stderr, "error: '%s' open failed\r\n", name );
         		return -1;
			}
        }else{
            (void) fprintf( stderr, "error: stat() fails for '%s'\r\n", name );
        	return -1;
        }
    }else if (S_ISREG(so.st_mode)){
		/* file exist => overwrite ? */		
		while(c != 'y' && c != 'n'){
			printf("%s is already existing, overwrite (y/n) ? ", name);

			c = 'y'; /* getchar(); fix this */
			printf("\n");

			/*
				#include <termios.h>
				...
				struct termios t0, t1;
				tcgetattr(0, &t0);
				...
				t1.c_lflag &= ~ECHO; // echo off
				t1.c_lflag &= ~ICANON; // one char at a time
			*/
		}
		if(c == 'y')
		{
      		if ((*fd = fopen(name, "wb")) == NULL)
      		{
         		(void) fprintf( stderr, "error: '%s' open failed\r\n", name );
         		return -1;
			}
		}else{
			return -1;
		}
	}

    return 0;   /* no error */
}

